#include "CrispBis.h"
#include "Vertex.h"
#include "Edge.h"
#include "Block.h"
#include "BlockEdge.h"
#include "Partition.h"
#include "SuperPartition.h"

#include <map>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <sstream>

namespace crispBis {

CrispBis::CrispBis(bool verbose) {
  this->verbose = verbose;
  emptyBlockEdges = NEW(DLList<BlockEdge>, DLList<BlockEdge>(true));
  nonemptyBlockEdges = NEW(DLList<BlockEdge>, DLList<BlockEdge>(true));
};

CrispBis::~CrispBis() {
  DELETE(emptyBlockEdges);
  DELETE(nonemptyBlockEdges);
  for (auto &it: Q)
    DELETE(it.second);
}

void CrispBis::initialize() {
  std::map<size_t, SuperBlock*> msb;

  for (auto &x: edgeLabels.idToName) {
    size_t r = x.first;
    Q[r] = NEW(SuperPartition,SuperPartition());
    msb[r] = NEW(SuperBlock, SuperBlock(Q[r]));
  }

  std::map<std::pair<Vertex*, size_t>, BlockEdge*> blockEdges;
  for (auto &x0: vertices.ptrToName) {
    Vertex *x = x0.first;
    for (auto &r0: edgeLabels.idToName) {
      size_t r = r0.first;
      blockEdges[std::make_pair(x,r)] = NEW(BlockEdge, BlockEdge(this));
    }
  }

  for (auto &x0: vertices.ptrToName) {
    Vertex *x = x0.first;
    for (auto &r0: edgeLabels.idToName) {
      size_t r = r0.first;
      for (Edge *e: x->comingEdges[r]) {
        e->blockEdge = blockEdges[std::make_pair(e->origin, e->label)];
        e->blockEdge->pushKey(e->degree);
      }
    }
  }

#ifndef COUNTING_SUCCESSORS
  std::map<std::pair<std::map<size_t, float>, std::map<size_t, float>>,
           DLList<Vertex>*> blocks;
#else
  std::map<std::pair<std::map<size_t, float>, std::map<size_t, BlockEdge>>,
           DLList<Vertex>*> blocks;
#endif // COUNTING_SUCCESSORS

  for (auto &x0: vertices.ptrToName) {
    Vertex *x = x0.first;
#ifndef COUNTING_SUCCESSORS
    std::map<size_t, float> tmp;
#else
    std::map<size_t, BlockEdge> tmp;
#endif // COUNTING_SUCCESSORS
    for (auto &r0: edgeLabels.idToName) {
      size_t r = r0.first;
#ifndef COUNTING_SUCCESSORS
      tmp[r] = blockEdges[std::make_pair(x,r)]->maxKey();
#else
      tmp[r] = *(blockEdges[std::make_pair(x,r)]);
#endif // COUNTING_SUCCESSORS
    }
    auto para = std::make_pair(x->label, tmp);
    auto it = blocks.find(para);
    if (it == blocks.end()) {
      DLList<Vertex> *lista = NEW(DLList<Vertex>, DLList<Vertex>(false));
      blocks[para] = lista;
      lista->add(x);
    }
    else
      it->second->add(x);
  }

  for (auto &e: blocks)
    NEW(Block, Block(e.second, &P, msb));

  emptyBlockEdges->clear();
};

void CrispBis::split(Block *X, SuperBlock *Y, size_t r) {
  std::vector<Vertex*> vertices_of_X;
  Vertex *x = X->vertices->head;
  while (x) {
    vertices_of_X.push_back(x);
    x = x->next;
  }

  computeBlockEdges(vertices_of_X, r);
  computeSubblocks(vertices_of_X, r);
  doSplitting(X, vertices_of_X, Y, r);
  clearAuxiliaryInfo(vertices_of_X, r);
}

void CrispBis::computeBlockEdges(
    std::vector<Vertex*> &vertices_of_X, size_t r)
{
  for (Vertex *x: vertices_of_X) {
    for (Edge *e: x->comingEdges[r]) {
      BlockEdge *bE = e->blockEdge;
      if (bE->departingBlockEdge == nullptr)
        bE->departingBlockEdge = NEW(BlockEdge, BlockEdge(this, bE));
      BlockEdge *dbE = bE->departingBlockEdge;
      bE->popKey(e->degree);
      dbE->pushKey(e->degree);
    }
  }
}

void CrispBis::computeSubblocks(
    std::vector<Vertex*> &vertices_of_X, size_t r)
{
  for (Vertex *x: vertices_of_X) {
    for (Edge *e: x->comingEdges[r]) {
      Vertex *v = e->origin;
      Block *bv = v->block;
      if (v->processed)
        continue;

      BlockEdge *bE = e->blockEdge;
      bv->vertices->remove(v);

#ifndef COUNTING_SUCCESSORS
      BlockEdge *dbE = bE->departingBlockEdge;
      float d1 = bE->maxKey();
      float d2 = dbE->maxKey();
      if (d1 >= d2) {
        auto it = bv->departingSubblocks2.find(d2);
        if (it == bv->departingSubblocks2.end()) {
          auto tmp = NEW(DLList<Vertex>, DLList<Vertex>(false));
          bv->departingSubblocks2[d2] = tmp;
          tmp->add(v);
        }
        else
          it->second->add(v);
      }
      else {
        auto it = bv->departingSubblocks1.find(d1);
        if (it == bv->departingSubblocks1.end()) {
          auto tmp = NEW(DLList<Vertex>, DLList<Vertex>(false));
          bv->departingSubblocks1[d1] = tmp;
          tmp->add(v);
        }
        else
          it->second->add(v);
      }
#else
      auto it = bv->departingSubblocks.find(*bE);
      if (it == bv->departingSubblocks.end()) {
        auto tmp = NEW(DLList<Vertex>, DLList<Vertex>(false));
        bv->departingSubblocks[*bE] = tmp;
        tmp->add(v);
      }
      else
        it->second->add(v);
#endif // COUNTING_SUCCESSORS

      v->processed = true;
    } // for e
  } // for x
}

void CrispBis::doSplitting(Block *X, std::vector<Vertex*> &vertices_of_X,
                           SuperBlock *Y, size_t r)
{
  Y->removeBlock(X);
  SuperBlock::createSuperBlock(Y->superPartition, X, r);

  for (Vertex *x: vertices_of_X) {
    for (Edge *e: x->comingEdges[r]) {
      if (e->blockEdge->departingBlockEdge)
        e->blockEdge = e->blockEdge->departingBlockEdge;
      assert(e->blockEdge);
      Vertex *v = e->origin;
      Block *bv = v->block;

#ifndef COUNTING_SUCCESSORS
      if (bv->departingSubblocks1.empty() &&
          bv->departingSubblocks2.empty()) {
        continue;
      }
#else
      if (bv->departingSubblocks.empty()) {
        continue;
      }
#endif // COUNTING_SUCCESSORS

      if (bv->vertices->empty()) {
        DELETE(bv->vertices);
#ifndef COUNTING_SUCCESSORS
        if (!bv->departingSubblocks1.empty()) {
          auto it = bv->departingSubblocks1.begin();
          bv->vertices = it->second;
          bv->departingSubblocks1.erase(it);
        }
        else {
          auto it = bv->departingSubblocks2.begin();
          bv->vertices = it->second;
          bv->departingSubblocks2.erase(it);
        }
#else
        auto it = bv->departingSubblocks.begin();
        bv->vertices = it->second;
        bv->departingSubblocks.erase(it);
#endif // COUNTING_SUCCESSORS
      }

#ifndef COUNTING_SUCCESSORS
      for (auto &it: bv->departingSubblocks1)
        Block::createBlock(it.second);
      bv->departingSubblocks1.clear();

      for (auto &it: bv->departingSubblocks2)
        Block::createBlock(it.second);
      bv->departingSubblocks2.clear();
#else
      for (auto &it: bv->departingSubblocks)
        Block::createBlock(it.second);
      bv->departingSubblocks.clear();
#endif // COUNTING_SUCCESSORS
    } // for e
  } // for x
}

void CrispBis::clearAuxiliaryInfo(
    std::vector<Vertex*> &vertices_of_X, size_t r)
{
  for (Vertex *x: vertices_of_X) {
    for (Edge *e: x->comingEdges[r]) {
      e->origin->processed = false;
      BlockEdge *bE = e->blockEdge;
      BlockEdge *sbE = bE->sourceBlockEdge;
      if (sbE) {
        sbE->departingBlockEdge = nullptr;
        bE->sourceBlockEdge = nullptr;
      }
    }
  }
  emptyBlockEdges->clear();
}

void CrispBis::compCB() {
  initialize();
  if (P.size() == 1)
    return;

  if (verbose)
    std::cerr << "P: " << P.str() << "\n";

  bool changed = true;
  while (changed) {
    changed = false;
    for (auto &r0: edgeLabels.idToName) {
      size_t r = r0.first;
      while (!Q[r]->compoundSuperBlocks->empty()) {
        SuperBlock *Y = Q[r]->compoundSuperBlocks->head;
        Block *X = Y->smallerBlock();
        split(X, Y, r);
        changed = true;
        if (verbose) {
          std::cerr << "P: " << P.str() << "\n";
          std::cerr << "Q[" << r0.second << "]: " << Q[r]->str() << "\n";
        }
      } // while
    } // for
  } // while
}

void CrispBis::addVertexLabel(const std::string ID,
                              const std::string p, float d) {
  Vertex *v = Vertex::getVertex(this, ID);
  v->addLabel(p, d);
}

void CrispBis::addEdge(const std::string xID, const std::string yID,
                       const std::string _r, float d) {
  size_t r = edgeLabels.getID(_r);
  Vertex *x = Vertex::getVertex(this, xID);
  Vertex *y = Vertex::getVertex(this, yID);
  NEW(Edge, Edge(r, x, y, d));
}

void CrispBis::read(const char *filename) {
  FILE *f;
  if (!filename)
    f = stdin;
  else if (!(f = fopen(filename, "r")))
      throw "Cannot open the input file.";

  char line[256];
  while (fgets(line, 256, f)) {
    std::stringstream iss(line);
    std::vector<std::string> l;
    std::string word;
    while (iss >> word)
      l.push_back(word);

    if (l.size() == 3)
      this->addVertexLabel(l[0], l[1], stof(l[2]));
    else if (l.size() == 4)
      this->addEdge(l[0], l[1], l[2], stof(l[3]));
    else if (l.size() == 0)
      continue;
    else {
      char buf[512];
      sprintf(buf, "Bad input line: %s", line);
      throw buf;
    }
  }

  if (filename)
    fclose(f);
}

void CrispBis::printGraph() {
  for (auto &it: vertices.ptrToName) {
    Vertex *v = it.first;
    std::cout << v->str() << "\n";
    for (auto &it: v->comingEdges)
      for (Edge *e: it.second)
        std::cout << e->str() << "\n";
  }
}

void CrispBis::main(bool verbose) {
  CrispBis *inst = NEW(CrispBis, CrispBis(verbose));
  inst->read();
  inst->compCB();
  std::cout << inst->P.str() << "\n";
  DELETE(inst);
  CHECK_MEMORY();
}

} // namespace crispBis

